/*
Copyright (C) 1996 Evgenii Rudnyi, http://Evgenii.Rudnyi.Ru/

Software for paper 
V. V. Kuzmenko, I. A. Uspenskaya, E. B. Rudnyi. Simultaneous assessment of calcium aluminates.  Bull. Soc. Chim. Belg. 1997, v. 106, N5, p. 235-243. 

This software is a copyrighted work licensed under the terms, described
in the file "FREE_LICENSE". 
*/

#include <iomanip.h>
#include "sumsqr.h"

void models();

char* LICENSE = "\
Evgenii Rudnyi 1996, (c) All rights reserved \n\n\
Chemistry Department \n\
Moscow State University \n\
119899 Moscow, Russia \n\
tel. (095)939-5452 \n\
rudnyi@mch.chem.msu.su \n\n\
This program is freeware (public domain). Feel free to use and \n\
distribute it, provided no charge is taken. \n\n\
I will be glad if you like this program. Let me know if you find any \n\
bugs. I would also appreciate your comments. \n\n\
If you need to change the model - try to contact me. \n\n\
Disclaimer of warranty: \n\
This program is supplied as is. I disclaim all warranties, \n\
express or implied, including, without limitation, the warranties of \n\
merchantability and of fitness of this program for any purpose. I assume \n\
no liability for damages direct or consequential, which may result from \n\
the use of this program.";

int nparam = 8;
param par[8];
static double& H1_298 = par[0].x;
static double& S1_298 = par[1].x;
static double& H2_298 = par[2].x;
static double& S2_298 = par[3].x;
static double& H3_298 = par[4].x;
static double& S3_298 = par[5].x;
static double& H4_298 = par[6].x;
static double& S4_298 = par[7].x;

static double F = 96.520;
static double H5_298 = 0.;
static double H6_298 = 0.;
static double S5_298 = 38.100;
static double S6_298 = 50.920;

static double Hf5_298 = -635090.;
static double Hf6_298 = -1675700.;

double Tr = 298.15;

double H5(double T)
{
  return
    H5_298+51.858*(T-Tr)+12.195e-4*(T*T-Tr*Tr)+2*4.683e5*(1./T-1./Tr);
}

double H6(double T)
{
  if (T == Tr)
    return H6_298;
  else if (T < 1200.01)
    return (H6_298+19165+122.679*(T-500)+3.439e-3*(T*T-500.*500.)
            +2*2.51185e6*(1./T-1./500.));
  else 
    return (H6_298+103272+107.571*(T-1200.)
            +8.2845e-3*(T*T-1200.*1200.));
}

double S5(double T)
{
  return (S5_298+51.858*(log(T)-log(Tr))+2*12.195e-4*(T-Tr)
          +4.683e5*(1./T/T-1./Tr/Tr));
}

double S6(double T)
{
  if (T == Tr)
    return S6_298;
  else if (T<1200.01)
    return (S6_298+48.502+122.679*(log(T)-log(500.))
            +2*3.439e-3*(T-500)+2.51185e6*(1./T/T-1./500./500.));
  else
    return (S6_298+152.414+107.571
            *(log(T)-log(1200.))+2*8.2845e-3*(T-1200));
}

double H1(double T)
{
return H1_298+1.*(H5(T) - H5_298) + 6.*(H6(T) - H6_298);
//return H1_298+1./7.*(H5(T) - H5_298) + 6./7.*(H6(T) - H6_298);
}

double H2(double T)
{
return
(H2_298+4.184*(66.09*(T-Tr)+2.74e-3*(T*T-Tr*Tr)+17.8e5*(1/T-1/Tr)));
}

double H3(double T)
{
return (H3_298+4.184*(36*(T-Tr)+2.98e-3*(T*T-Tr*Tr)+7.96e5*(1/T-1/Tr)));
}

double H4(double T)
{
return
(H4_298+4.184*(62.28*(T-Tr)+2.29e-3*(T*T-Tr*Tr)+12.01e5*(1/T-1/Tr)));
}

double S1(double T)
{
//return S1_298+1./7.*(S5(T) - S5_298) + 6./7.*(S6(T) - S6_298);
return S1_298+1.*(S5(T) - S5_298) + 6.*(S6(T) - S6_298);
}

double S2(double T)
{
return (S2_298+4.184*(66.09*(log(T)-log(Tr))+2*2.74e-3*(T-Tr)+
        17.8e5/2.*(1./T/T-1./Tr/Tr)));
}

double S3(double T)
{
return (S3_298+4.184*(36*(log(T)-log(Tr))+2*2.98e-3*(T-Tr)+
        7.96e5/2.*(1./T/T-1./Tr/Tr)));
}

double S4(double T)
{
return (S4_298+4.184*(62.28*(log(T)-log(Tr))+2*2.29e-3*(T-Tr)+
        12.01e5/2.*(1./T/T-1./Tr/Tr)));
}

double eq1(double *reg)
{
  double& T = reg[1];
  double& E = reg[0];
  return E - (-1./F/2.*(H1(T)-H5(T)-6*H6(T)-T*(S1(T)-S5(T)-6*S6(T))));
}

double eq2(double *reg)
{
  double& T = reg[1];
  double& E = reg[0];
  return E - (-1./F/4.*(3*H2(T)-H1(T)-2*H5(T)-T*(3*S2(T)-S1(T)-2*S5(T))));
}

double eq3(double *reg)
{
  double& T = reg[1];
  double& E = reg[0];
  return E - (-1./F/2.*(2*H3(T)-H5(T)-H2(T)-T*(2*S3(T)-S5(T)-S2(T))));
}

double eq4(double *reg)
{
  double& T = reg[1];
  double& E = reg[0];
  return E - (-1./F/4.*(H4(T)-H3(T)-2*H5(T)-T*(S4(T)-S3(T)-2*S5(T))));
}

double eq5(double *reg)
{
  double& T = reg[1];
  double& E = reg[0];
  return E - (-1./F/(4./3.)*(H1(T)-H2(T)-4*H6(T)-T*(S1(T)-S2(T)-4*S6(T))));
}

double eq6(double *reg)
{
  double& T = reg[1];
  double& E = reg[0];
  return E - (-1./F/4.*(-H1(T)+5*H2(T)-4*H3(T)-T*(-S1(T)+5*S2(T)-4*S3(T))));
}

double eq7(double *reg)
{
  double& T = reg[1];
  double& E = reg[0];
  return E - (-1./F/4.*(5*H3(T)-2*H2(T)-H4(T)-T*(5*S3(T)-2*S2(T)-S4(T))));
}

double S(double *reg)
{
  double& S = reg[0];
  double& x = reg[1];
  if (x == 0.8571)
    return (S - S1_298)/7.;
  else if (x == 0.6667)
    return (S - S2_298)/3.;
  else if (x == 0.5000)
    return (S - S3_298)/2.;
  else if (x == 0.2500)
    return (S - S4_298)/4.;
  else
  {
    cout << "error in function S" << endl;
    exit(1);
    return 0.;
  }
}

double H(double *reg)
{
  double& H = reg[0];
  double& x = reg[1];
  double& T = reg[2];
  if (x == 0.8571)
    return (H - (H1(T)-H5(T)-6*H6(T)))/1e3/7.;
  else if (x == 0.6667)
    return (H - (H2(T)-H5(T)-2*H6(T)))/1e3/3.;
  else if (x == 0.5000)
    return (H - (H3(T)-H5(T)-H6(T)))/1e3/2.;
  else if (x == 0.2500)
    return (H - (H4(T)-3*H5(T)-H6(T)))/1e3/4.;
  else
  {
    cout << "error in function H" << endl;
    exit(1);
    return 0.;
  }
}

void init_res()
{
  res.insert(residual("eq1", eq1, 2, 0x3));
  res.insert(residual("eq2", eq2, 2, 0xF));
  res.insert(residual("eq3", eq3, 2, 0x3C));
  res.insert(residual("eq4", eq4, 2, 0xF0));
  res.insert(residual("eq5", eq5, 2, 0xF));
  res.insert(residual("eq6", eq6, 2, 0x3F));
  res.insert(residual("eq7", eq7, 2, 0xFC));
  res.insert(residual("S", S, 2, 0xAA));
  res.insert(residual("H", H, 3, 0x55));
}

char* EQ = "";

void init_par()
{
  par[0].name = "H1_298";
  par[0].atr = true;
  par[1].name = "S1_298";
  par[1].atr = true;
  par[2].name = "H2_298";
  par[2].atr = true;
  par[3].name = "S2_298";
  par[3].atr = true;
  par[4].name = "H3_298";
  par[4].atr = true;
  par[5].name = "S3_298";
  par[5].atr = true;
  par[6].name = "H4_298";
  par[6].atr = true;
  par[7].name = "S4_298";
  par[7].atr = true;
}

void init_est()
{
}

void models()
{
//  memory test;
  ofstream of1((file_name + "axm").c_str());
  of1.precision(prcsn);
  of1.setf(ios::showpoint);
  of1 << "T E1 E2 E3 E4 E5 E6 E7 " << endl;
  double prs[2];
  prs[0] = 0.;
  for (prs[1] = 900.; prs[1] <= 1501.; prs[1] += 25.)
  {
    double tmp;
    of1 << setw(wdth) << prs[1];
    tmp = -eq1(prs);
    of1 << setw(wdth) << tmp;
    tmp = -eq2(prs);
    of1 << setw(wdth) << tmp;
    tmp = -eq3(prs);
    of1 << setw(wdth) << tmp;
    tmp = -eq4(prs);
    of1 << setw(wdth) << tmp;
    tmp = -eq5(prs);
    of1 << setw(wdth) << tmp;
    tmp = -eq6(prs);
    of1 << setw(wdth) << tmp;
    tmp = -eq7(prs);
    of1 << setw(wdth) << tmp;
    of1 << endl;
  }
}

void deviates()
{
  ofstream of1((file_name + "ax1").c_str());
  of1.precision(prcsn);
  of1.setf(ios::showpoint);
  iser = ser.begin();
  while(iser != ser.end())
  {
    series& t = *iser;
    of1 << setw(10 - t.ID.stringlen()) << 'T' << t.ID;
    of1 << setw(10) << t.ID;
    iser++;
  }
  of1 << endl;
  int end = 0;
  for (int i = 0; !end; ++i)
  {
    end = 1;
    iser = ser.begin();
    while(iser != ser.begin())
    {
      series& t = *iser;
      if (i < t.Ntot())
      {
        end = 0;
        of1 << setw(10) << t.val(i, 1);
        of1 << setw(10) << (*t.f)(t(i));
      }
      else
      {
        of1 << setw(10) << "miss";
        of1 << setw(10) << "miss";
      }
      iser++;
    }
    of1 << endl;
  }
  of << endl;
}

void sb_vs_sa()
{
  ofstream of2((file_name + "ax2").c_str());
  of2.precision(prcsn);
  of2.setf(ios::showpoint);
  of2 << " ID1 err_a1 err_b1 ID2 err_a2 err_b2 ID3 err_a3 err_b3" << endl;

  struct err
  {
    cString ID;
    double err_a;
    double err_b;
    err() : ID(), err_a(0.), err_b(0.) {}
  };
  err errors[20][3];

  iser = ser.begin();
  int i1 = 0, i2 = 0, i3 = 0;
  while(iser != ser.end())
  {
    series& t = *iser;
    if (t.name(0) == "E")
    {
      if (t.hide)
      {
        set_sums_1(t);
        errors[i1][0].ID = "*";
      }
      errors[i1][0].ID += t.ID;
      errors[i1][0].err_a = t.suma/t.Ns()/3.8;
      if (t.Ps())
        errors[i1][0].err_b = t.sumb/t.Ps()*sqrt(t.Ps()/t.Ns())/3.8;
      else
        errors[i1][0].err_b = 0.;
      i1++;
    }
    else if (t.name(0) == "S_298")
    {
      if (t.hide)
      {
        set_sums_1(t);
        errors[i2][1].ID = "*";
      }
      errors[i2][1].ID += t.ID;
      errors[i2][1].err_a = t.suma/t.Ns()/.4;
      if (t.Ps())
        errors[i2][1].err_b = t.sumb/t.Ps()*sqrt(t.Ps()/t.Ns())/.4;
      else
        errors[i2][1].err_b = 0.;
      i2++;
    }
    else if (t.name(0) == "H")
    {
      if (t.hide)
      {
        set_sums_1(t);
        errors[i3][2].ID = "*";
      }
      errors[i3][2].ID += t.ID;
      errors[i3][2].err_a = t.suma/t.Ns()/1.;
      if (t.Ps())
        errors[i3][2].err_b = t.sumb/t.Ps()*sqrt(t.Ps()/t.Ns())/1.;
      else
        errors[i3][2].err_b = 0.;
      i3++;
    }
    iser++;
  }
  for (int i = 0; i < i1; ++i)
  {
    for (int j = 0; j < 3; ++j)
    {
      if (errors[i][j].ID != "")
        of2 << setw(wdth) << errors[i][j].ID
            << setw(wdth) << errors[i][j].err_a
            << setw(wdth) << errors[i][j].err_b;
      else
        of2 << setw(wdth) << "miss"
            << setw(wdth) << "miss"
            << setw(wdth) << "miss";
   }
   of2 << endl;
  }
}

void H_and_S()
{
  ofstream of3((file_name + "ax3").c_str());
  of3.precision(prcsn);
  of3.setf(ios::showpoint);
  of3 << " x    H  sH   S   sS  G1  G2  G3" << endl;
  double x[4] = {6./7., 2./3., .5, 0.25};
  double nm[4] = {7., 3., 2., 4.};
  double (*fH[4])(double T) = {H1, H2, H3, H4};
  double (*fS[4])(double T) = {S1, S2, S3, S4};
  of3 << setw(wdth) << 1.
      << setw(wdth) << 0.
      << setw(wdth) << 0.
      << setw(wdth) << 0.
      << setw(wdth) << 0.
      << setw(wdth) << 0.
      << setw(wdth) << 0.
      << setw(wdth) << 0.
      << endl;
  double T1 = 298.15, T2 = 1300., T3 = 2000.;
  for (int i = 0; i < 4; ++i)
    of3 << setw(wdth) << x[i]
        << setw(wdth) << (par[2*i].x/nm[i] - H5_298*(1. - x[i]) - H6_298*x[i])/1000.
        << setw(wdth) << (par[2*i].sx/nm[i])/1000.
        << setw(wdth) << par[2*i + 1].x/nm[i] - S5_298*(1. - x[i]) - S6_298*x[i]
        << setw(wdth) << par[2*i + 1].sx/nm[i]
        << setw(wdth) << ((fH[i](T1) - T1*fS[i](T1))/nm[i]
                         - (H5(T1) - T1*S5(T1))*(1. - x[i])
                         - (H6(T1) - T1*S6(T1))*x[i])/1000.
        << setw(wdth) << ((fH[i](T2) - T2*fS[i](T2))/nm[i]
                         - (H5(T2) - T2*S5(T2))*(1. - x[i])
                         - (H6(T2) - T2*S6(T2))*x[i])/1000.
        << setw(wdth) << ((fH[i](T3) - T3*fS[i](T3))/nm[i]
                         - (H5(T3) - T3*S5(T3))*(1. - x[i])
                         - (H6(T3) - T3*S6(T3))*x[i])/1000.
        << endl;
  of3 << setw(wdth) << 0.
      << setw(wdth) << 0.
      << setw(wdth) << 0.
      << setw(wdth) << 0.
      << setw(wdth) << 0.
      << setw(wdth) << 0.
      << setw(wdth) << 0.
      << setw(wdth) << 0.
      << endl;
}

void post_analysis()
{
  of << endl << "******     post analysis     *****" << endl;
  double T = 1063.;
  of << "CA6 "
     << ((H1(T)-H5(T)-6*H6(T)) - (H1(Tr)-H5(Tr)-6*H6(Tr)))/1e3
     << endl;
  of << "CA2 "
     << ((H2(T)-H5(T)-2*H6(T)) - (H2(Tr)-H5(Tr)-2*H6(Tr)))/1e3
     << endl;
  of << "CA  "
     << ((H3(T)-H5(T)-H6(T)) - (H3(Tr)-H5(Tr)-H6(Tr)))/1e3
     << endl;
  of << "C3A "
     << ((H4(T)-3.*H5(T)-H6(T)) - (H4(Tr)-3.*H5(Tr)-H6(Tr)))/1e3
     << endl;
  of << endl;
  models();
  deviates();
  sb_vs_sa();
  H_and_S();
}

/*
void print_G(double T)
{
  cout << setw(10) << H6(T) - H6_298 << setw(10) << S6(T) - S6_298 << endl;
  cout << setw(10) << H5(T) - H5_298 << setw(10) << S5(T) - S5_298 << endl;
  cout << setw(10) << (H2(T) - H2_298)/4.184
       << setw(10) << (S2(T) - S2_298)/4.184 << endl;
  cout << setw(10) << (H3(T) - H3_298)/4.184
       << setw(10) << (S3(T) - S3_298)/4.184 << endl;
  cout << setw(10) << (H4(T) - H4_298)/4.184
       << setw(10) << (S4(T) - S4_298)/4.184 << endl;
  cout << endl;
}

int main()
{
  double T = 500.;
  print_G(T);
  T = 900.;
  print_G(T);
  T = 1200.;
  print_G(T);
  T = 1500.;
  print_G(T);
  return 0;
}
*/

