/*
Copyright (C) 1995 Evgenii Rudnyi, http://Evgenii.Rudnyi.Ru

Software for paper 
E.B. Rudnyi. Statistical model of systematic errors: linear error model. Chemometrics and Intelligent Laboratory Systems. 1996, V. 34, N 1, p. 41-54. 

This software is a copyrighted work licensed under the terms, described
in the file "FREE_LICENSE". 
*/

#include "sumsqr.h"
#include "iomanip.h"

char* LICENSE = "\
Evgenii Rudnyi 1995, (c) All rights reserved \n\n\
Chemistry Department \n\
Moscow State University \n\
119899 Moscow, Russia \n\
tel. (095)939-5452 \n\
rudnyi@comp.chem.msu.su \n\n\
This program is freeware (public domain). Feel free to use and \n\
distribute it, provided no charge is taken. \n\n\
I will be glad if you like this program. Let me know if you find any \n\
bugs. I would also appreciate your comments. \n\n\
If you need to change the model - try to contact me. \n\n\
Disclaimer of warranty: \n\
This program is supplied as is. I disclaim all warranties, \n\
express or implied, including, without limitation, the warranties of \n\
merchantability and of fitness of this program for any purpose. I assume \n\
no liability for damages direct or consequential, which may result from \n\
the use of this program.";

extern "C"
int hscor_(double *tf, double *dh1, double *ds1, double *dh2, double *ds2);

int nparam = 4;
param par[4];

static double& H1 = par[0].x;
static double& S1 = par[1].x;
static double& H2 = par[2].x;
static double& S2 = par[3].x;
static double R = 8.31441;

static double dh1;
static double ds1;
static double dh2;
static double ds2;

double ptot(double *reg)
{
  double& lnp = reg[0];
  double T = 1e3/reg[1];
  hscor_(&T, &dh1, &ds1, &dh2, &ds2);
  double lnpm = - (H1 + dh1)/R/T + (S1 + ds1)/R;
  double lnpdm = - (H2 + dh2)/R/T + (S2 + ds2)/R;
  return lnp - (lnpm + log(1. + exp(lnpdm)));
}

double pKE(double *reg)
{
  double& lnp = reg[0];
  double T = 1e3/reg[1];
  hscor_(&T, &dh1, &ds1, &dh2, &ds2);
  double lnpm = - (H1 + dh1)/R/T + (S1 + ds1)/R;
  double lnpdm = - (H2 + dh2)/R/T + (S2 + ds2)/R;
  return lnp - (lnpm + log(1. + sqrt(2.)*exp(lnpdm)));
}

double pTR(double *reg)
{
  double& lnp = reg[0];
  double T = 1e3/reg[1];
  double& psys = reg[2];
  hscor_(&T, &dh1, &ds1, &dh2, &ds2);
  double lnpm = - (H1 + dh1)/R/T + (S1 + ds1)/R;
  double lnpdm = - (H2 + dh2)/R/T + (S2 + ds2)/R;
  return lnp - (lnpm + log(1. + 2.*exp(lnpdm))
         - log(1. + exp(lnpm)*exp(lnpdm)/psys));
}


void init_res()
{
  res.insert(residual("ptot", ptot, 2, 0xF));
  res.insert(residual("pKE",  pKE,  2, 0xF));
  res.insert(residual("pTR", pTR, 3, 0xF));
}

char* EQ = "\
There are equations as follows \n\
\t\t\teq_ID               equation \n\n\
\t\t\tptot                total pressure\n\
\t\t\tpKE                 Knudsen effusion\n\
\t\t\tpTR                 transpiration\n\n\
available in this release, \n\
and four unknown parameters: \n\
DelH and DelS of two reactions at 1044 K \n\
\tKCl(cd) = KCl(g) \n\
\tKCl(cd) + KCl(g) = K2Cl2(g) \n";

void init_par()
{
  par[0].name = "H1";
  par[0].atr = true;
  par[1].name = "S1";
  par[1].atr = true;
  par[2].name = "H2";
  par[2].atr = true;
  par[3].name = "S2";
  par[3].atr = true;
}

void init_est()
{
}

void post_analysis()
{
  double H1best = 206775.;
  double S1best = 133.03;
  double H2best = 37834.;
  double S2best = 24.913;
  of << endl << "******     post analysis     *****" << endl;
  of << "invT     diftot     difKE      difTR      pdm" << endl;
  for (double invT = .55; invT < 1.62; invT += 0.02)
  {
    double T = 1e3/invT;
    hscor_(&T, &dh1, &ds1, &dh2, &ds2);
    double pm = - (H1 + dh1)/R/T + (S1 + ds1)/R;
    double pdm = - (H2 + dh2)/R/T + (S2 + ds2)/R;
    double ptot = pm + log(1. + exp(pdm));
    double pKE = pm + log(1. + sqrt(2.)*exp(pdm));
    double pTR = pm + log(1. + 2.*exp(pdm))
                  - log(1. + exp(pm)*exp(pdm));
    double pm1 = - (H1best + dh1)/R/T + (S1best + ds1)/R;
    double pdm1 = - (H2best + dh2)/R/T + (S2best + ds2)/R;
    double ptot1 = pm1 + log(1. + exp(pdm1));
    double pKE1 = pm1 + log(1. + sqrt(2.)*exp(pdm1));
    double pTR1 = pm1 + log(1. + 2.*exp(pdm1))
                   - log(1. + exp(pm1)*exp(pdm1));
    of << setw(wdth) << invT
       << setw(wdth) << ptot - ptot1
       << setw(wdth) << pKE - pKE1
       << setw(wdth) << pTR - pTR1
       << setw(wdth) << pdm
       << endl;
  }
}

