/*
Copyright (C) 1994-1998 Evgenii Rudnyi, http://Evgenii.Rudnyi.Ru/

Implementation of the algorithm described in  
E.B. Rudnyi. Statistical model of systematic errors: linear error model. Chemometrics and Intelligent Laboratory Systems. 1996, V. 34, N 1, p. 41-54. 

This software is a copyrighted work licensed under the terms, described
in the file "FREE_LICENSE". 
*/

#include <math.h>
#include <float.h>
#include <ctype.h>
#include "sumsqr.h"

static const char* HELP ="\n\
Evgenii Rudnyi 1994, 1995, (c) All rights reserved \n\
version 1.51 (15 September 1995) \n\
processing experimental series with systematic errors \n\
see file README.TXT for more information \n\n\
    program [-options] datafile [setfile] \n\
           options \n\
    -o  output file (default - setfile.lst or datafile.lst) \n\
    -h  help \n\
    -l  license \n\
    -p  print results only (no iterations results) (default) \n\
    -p1 print the results of big iterations \n\
    -p2 print all the intermediate resutls \n\
    -s  show used parameters (default) \n\
    -s1 show free parameters only \n\
    -s2 show all parameters \n\
    -e  listing without experimental values \n\
    -e1 print only used series \n\
    -e2 print all the series \n\
    -b  start with variance estimates \n\
    -b1 start with parameter estimates \n";

static const char* WRONGOPTION = "\nNo such option \n\n";
static const char* WRONGARG = "\nWrong set of arguments \n\n";
static const char* NOTEXIST = " - such file does not exist \n";
static const char* CANTCREATE = "\nCan not create a file - ";
static const char* NOEQ = "\nCan not find appropriate equation - ";
cString file_name;

istream& operator>>(istream& in, cString& to)
{
  char ch;
  to = "";
  in >> ws;
  while (ch = in.peek(), !isspace(ch) &&
                            ch != ',' &&
                            ch != ';' &&
                            ch != EOF)
  {
    ch = in.get();
    if (!Buffer.write(&ch))
    {
      Buffer.clear();
      return in;
    }
  }
  int len = Buffer.busy();
  if (!len) return in;
  if (!Buffer.write("\0"))
  {
    Buffer.clear();
    return in;
  }
  to = Buffer.ptr();
  Buffer.clear();
  return in;
}

boolean open_streams(int argc, char* argv[], ifstream& df, ifstream& sf)
{
  cString datafile, setfile, outfile;
  cString spoint(".");
  if (argc == 1)
  {
    cout << HELP << endl << "press RETURN to continue" << endl;
    cin.get();
    cout << EQ;
    return false;
  }
  for (int i = 1; i < argc; i++)
  {
    if (argv[i][0] == '-' || argv[i][0] == '/')
    {
      switch (tolower(argv[i][1]))
      {
        case 'o' :
          outfile = argv[i] + 2; break;
//          case 'a' :
//            sys_errors = absolute; break;
        case 'h' : case '?' :
          cout << HELP << endl << "press and key" << endl;
          cin.get();
          cout << EQ;
          return false;
        case 'l' :
          cout << LICENSE; return false;
        case 'p' :
          switch (argv[i][2])
          {
            case '\0': case '0':
              print = results; break;
            case '1':
              print = big_iter; break;
            case '2':
              print = all_iter; break;
          }
          break;
        case 's' :
          switch (argv[i][2])
          {
            case '\0': case '0':
              show = used; break;
            case '1':
              show = free_par; break;
            case '2':
              show = all_par; break;
          }
          break;
        case 'e' :
          switch (argv[i][2])
          {
            case '\0': case '0':
              print_series = no_series; break;
            case '1':
              print_series = used_series ; break;
            case '2':
              print_series = all_series ; break;
          }
          break;
        case 'b' :
          switch (argv[i][2])
          {
            case '\0': case '0':
              start = variances; break;
            case '1':
              start = parameters; break;
          }
          break;
        default:
          cout << WRONGOPTION << HELP ; return false;
      }
    }
    else if (!datafile.stringlen())
      datafile = argv[i];
    else if (!setfile.stringlen())
      setfile = argv[i];
    else
    {
      cout << WRONGARG << HELP ;
      return false;
    }
  }
  if (!datafile.stringlen())
  {
    cout << WRONGARG << HELP ;
    return false;
  }
  else
  {
    if (datafile.pos(spoint) == NO_SUBSTR_FOUND)
      datafile = datafile + ".dat";
    df.open(datafile.getstr(), ios::in);
    if (!df)
    {
      cout << endl << datafile << NOTEXIST;
      return false;
    }
  }
  if (setfile.stringlen())
  {
    if (setfile.pos(spoint) == NO_SUBSTR_FOUND)
      setfile = setfile + ".set";
    if (setfile != datafile)
    {
      sf.open(setfile.getstr(), ios::in);
      if (!sf)
      {
        cout << endl << setfile << NOTEXIST;
        return false;
      }
    }
    else
    {
      cout << "setfile has the same name as datafile" << endl;
      return false;
    }
  }
  else
  {
    setfile = datafile % "." + "set";
    sf.open(setfile.getstr(), ios::in);
  }
  if (!outfile.stringlen())
  {
    int pos;
    outfile = setfile;
    if (outfile == outfile % "." + "lst")
      outfile = outfile % "." + "ls1";
    else
      outfile = outfile % "." + "lst";
  }
  else
    if (outfile.pos(spoint) == NO_SUBSTR_FOUND)
      outfile = outfile + ".lst";
  file_name = outfile % ".";

  of.open(outfile.getstr());
  of.precision(prcsn);
  of.setf(ios::showpoint);
  if (!of)
  {
    cout << CANTCREATE << outfile << endl;
    return false;
  }
  return true;
}

void read_series(ifstream& df)
{
  char ch;
  int np;
  while (df)
  {
    series t;
    df >> ws;
    if (df.peek() == '*')
    {
      t.hide = true;
      df.ignore();
    }
    df >> t.ID;
    while (ch = df.get(), ch != ',' &&
                          ch != ';' &&
                          ch != EOF);
    if (ch != ',')
    {
      continue;
    }
    df >> t.eq;
    np = 0;
    ires = res.begin();
    while (ires != res.end() && t.eq >= *ires)
    {
      if (t.eq == *ires)
      {
        t.f = (*ires).getr();
        t.mask = (*ires).getmask();
        np = (*ires).getnp();
        break;
      }
      ires++;
    }
    if (!t.f || !t.mask || !np)
    {
      cout << NOEQ << t.ID << " - " << t.eq << endl;
      of << NOEQ << t.ID << " - " << t.eq << " -  series ignored" << endl;
      while (ch = df.get(), ch != ';' &&
                            ch != EOF);
      continue;
    }
    while (ch = df.get(), ch != ',' &&
                          ch != ';' &&
                          ch != EOF);
    if (ch != ',')
    {
      continue;
    }
    t.read(df, np);
    if (!t)
    {
      continue;
    }
    else
      ser.push_back(t);
  }
}

void read_var_(ifstream& sf, int& fl, double& var, double* same_var, int& i)
{
  i = 0;
  double tmp = HUGE_VAL;
  sf >> ws;
  if (sf.peek() == '*')
  {
    fl = series::fixed;
    sf.ignore();
  }
  else if (sf.peek() == '#')
  {
    fl = series::same;
    sf.ignore();
    if (isdigit(sf.peek()))
    {
      sf >> i;
      if (i >= nvar) i = 0;
      sf.clear();
    }
  }
  else if (sf.peek() == '%')
  {
    fl = series::own;
    sf.ignore();
  }
  sf >> tmp;
  if (tmp != HUGE_VAL)
  {
    var = tmp*tmp;
    if (fl == series::same)
      same_var[i] = var;
  }
  sf.clear();
}

char read_var(ifstream& sf, int& fl, double& var, double* same_var, int& i)
{
  read_var_(sf, fl, var, same_var, i);
  char ch;
  while (ch = sf.get(), ch != ',' &&
                             ch != ';' &&
                             ch != EOF);
  return ch;
}

char read_var_1(ifstream& sf, int& fl, double& var, double* same_var,
                                                int& i, double& adjust)
{
  read_var_(sf, fl, var, same_var, i);
  sf >> ws;
  if (sf.peek() == '@')
  {
    sf.ignore();
    double tmp = HUGE_VAL;
    sf >> tmp;
    if (tmp != HUGE_VAL)
      adjust = tmp*tmp;
    sf.clear();
  }
  char ch;
  while (ch = sf.get(), ch != ',' &&
                             ch != ';' &&
                             ch != EOF);
  return ch;
}

void read_setfile(ifstream& sf)
{
  cString word;
  boolean hide;
  char ch;
  series* tptr;
  int fl, j;
  while (sf >> ws, sf)
  {
    hide = false;
    word = "";
    if (sf.peek() == '*')
    {
      hide = true;
      sf.ignore();
    }
    sf >> word;
    if (word == "START")
      break;
    tptr = NULL;
    iser = ser.begin();
    while (iser != ser.end())
    {
      if (word == (*iser).ID)
      {
        tptr = &(*iser);
        tptr->hide = hide;
        break;
      }
      iser++;
    }
    if (!tptr)
    {
      of << "series " << word << " not found - ignored" << endl;
      cout << "series " << word << " not found - ignored" << endl;
      while (ch = sf.get(), ch != ';' &&
                            ch != EOF);
    }
    else
    {
      while (ch = sf.get(), ch != ',' &&
                            ch != ';' &&
                            ch != EOF);
      if (ch != ',')
        continue;
      if (read_var_1(sf, tptr->fl_sr2, tptr->sr2, sr2t, tptr->isr2,
                                                    tptr->adjust) != ',')
        continue;
      if (read_var(sf, tptr->fl_ga, tptr->ga, gat, tptr->igat) != ',')
        continue;
      if (read_var(sf, tptr->fl_gb, tptr->gb, gbt, tptr->igbt) != ',')
        continue;
      while (ch = sf.get(), ch != ';' &&
                            ch != EOF);
    }
  }
  if (sf)
  {
    ch = 0;
    while (sf >> ws, ch != ';' && sf.peek() != EOF)
    {
      hide = false;
      word = "";
      if (sf.peek() == '*')
      {
        hide = true;
        sf.ignore();
      }
      sf >> word;
      if (word != "")
      {
        for (j = 0; j < nparam; j++)
        {
          if (word == par[j].name)
          {
            if (hide)
              par[j].atr = false;
            else
              par[j].atr = true;
            sf >> par[j].x;
            sf.clear();
            break;
          }
        }
        if (j == nparam)
        {
          of << "parameter " << word << " not found - ignored" << endl;
          cout << "parameter " << word << " not found - ignored" << endl;
        }
      }
      while (ch = sf.get(), ch != ',' &&
                            ch != ';' &&
                            ch != EOF);
    }
  }
}

