/*
Copyright (C) 1994-1998 Evgenii Rudnyi, http://Evgenii.Rudnyi.Ru/

Implementation of the algorithm described in  
E.B. Rudnyi. Statistical model of systematic errors: linear error model. Chemometrics and Intelligent Laboratory Systems. 1996, V. 34, N 1, p. 41-54. 

This software is a copyrighted work licensed under the terms, described
in the file "FREE_LICENSE". 
*/

#include <iomanip.h>
#include <ctype.h>
#include <math.h>
#include "sumsqr.h"

ostream& operator<<(ostream& out, param& old)
{
  if (!old.atr) out << '*'; else out << ' ';
  return out << setw(wdth) << old.name << setw(wdth)
             << old.x;
}

void print_par(param& t, double s)
{
  switch (show)
  {
    case used :
      if (t.use)
      {
        of << t;
        if (s != HUGE_VAL)
          of << "  +-  " << setw(wdth)
             << s << endl;
        else
          of << endl;
      }
      break;
    case free_par :
      if (t.atr)
      {
        of << t;
        if (s != HUGE_VAL)
          of << "  +-  " << setw(wdth)
             << s << endl;
        else
          of << endl;
      }
      break;
    case all_par :
      of << t;
      if (s != HUGE_VAL)
        of << "  +-  " << setw(wdth)
           << s << endl;
      else
        of << endl;
      break;
    default:
      cout << "something wrong with variable show";
      break;
  }
}

void print_ss_par_1()
{
  double ssq = 0.;
  iser = ser.begin();
  while (iser != ser.end())
  {
    if (!(*iser).hide)
      ssq += (*iser).SS;
    iser++;
  }
  of << "SSQ " << setw(wdth) << ssq << endl;
  for (int j = 0; j < nparam; j++)
    print_par(par[j]);
}

void print_zxssq_1(int infer, int ier, double* work)
{
  of << "ZXSSQ  reports:" << endl;
  if (ier == 38)
    of << "jacobian is zero - the solution x is a stationary point"
           << endl;
  switch (infer)
  {
    case 0:
      of << "convergence failed - " << endl;
      switch (ier)
      {
        case 129:
          of << "\ta singularity was detected \
                 in the jacobian and recovery failed" << endl;
          break;
        case 130:
          of << "\tat least one of m, n, iopt, parm(1), or parm(2) \n\
                  was specified incorrectly" << endl;
          break;
        case 131:
          of << "\tthe marquardt parameter exceeded parm(3)" << endl;
          break;
        case 132:
          of << "\tafter a successful recovery from a singular jacobian, \n\
                 the vector x has cycled back to the first \
                 singularity" << endl;
          break;
        case 133:
          of << "\tmaxfn was exceeded" << endl;
          break;
        default:
          of << "\tit is not clear what has happened" << endl;
      }
      break;
    case 1:
      of << "criterion NSIG was satisfied" << endl;
      break;
    case 2:
      of << "criterion EPS was satisfied" << endl;
      break;
    case 3:
      of << "criterions NSIG and EPS were satisfied" << endl;
      break;
    case 4:
      of << "criterion DELTA was satisfied" << endl;
      break;
    case 5:
      of << "criterions NSIG and DELTA were satisfied" << endl;
      break;
    case 6:
      of << "criterions EPS and DELTA were satisfied" << endl;
      break;
    case 7:
      of << "criterions NSIG, EPS and DELTA were satisfied" << endl;
      break;
    default:
      of << "it is not clear what has happened" << endl;
  }
  of << "Gragient            " << setw(wdth)
     << work[0] << endl;
  of << "Function evaluations" << setw(wdth)
     << work[1] << endl;
  of << "Significant digits  " << setw(wdth)
     << work[2] << endl;
  of << "Marquardt parameter " << setw(wdth)
     << work[3] << endl;
  of << "Number iterations   " << setw(wdth)
     << work[4] << endl;
}

void print_all_points()
{
  double tmp;
  of << endl << "Results " << endl;
  of << " ID           eq            sr             sga            sgb" << endl;
  iser = ser.begin();
  while(iser != ser.end())
  {
    series& t = *iser;
    if (t.hide && (print_series == all_series))
    {
      of << '*';
      set_sums_1(t);
      if (t.Ns() > 1)
      {
        if (t.Ps())
        {
          if (t.Ns() == 2)
          {
            t.sr2 = t.sumi/2.;
            t.gb = 0.;
            t.ga = 0.;
          }
          else
          {
            t.sr2 = (t.sumi - t.suma*t.suma/t.Ns() - t.sumb*t.sumb/t.Ps())
                               /(t.Ns() - 2);
            t.gb = (t.sumb*t.sumb/t.Ps()/t.sr2 - 1)/t.Ps();
            t.ga = (t.suma*t.suma/t.Ns()/t.sr2 - 1)/t.Ns();
          }
        }
        else
        {
          t.sr2 = (t.sumi - t.suma*t.suma/t.Ns())/(t.Ns() - 1);
          t.gb = 0.;
          t.ga = (t.suma*t.suma/t.Ns()/t.sr2 - 1)/t.Ns();
        }
      }
      else
      {
        t.sr2 = t.sumi;
        t.ga = 0.;
        t.gb = 0.;
      }
      if (t.ga < 0.) t.ga = 0.;
      if (t.gb < 0.) t.gb = 0.;
    }
    else if (!t.hide)
      of << ' ';

    if (!t.hide || (print_series == all_series))
    {
      of << setw(6) << t.ID << setw(wdth+2) << t.eq << ", ";
      switch(t.fl_sr2)
      {
        case series::fixed:
          of << "*  ";
          break;
        case series::same:
          of << "#";
          if (t.isr2 == 0)
            of << "  ";
          else if (t.isr2 < 10)
            of << t.isr2 << " ";
          else
            of << t.isr2;
          break;
        case series::own:
          of << "%  ";
          break;
      }
      of << setw(wdth-2) << sqrt(t.sr2);
      if (t.adjust)
        of << setw(wdth-2) << sqrt(t.adjust) << ", ";
      else
        of << ", ";
//        of << setw(wdth + 2) << ", ";
      switch(t.fl_ga)
      {
        case series::fixed:
          of << "*  ";
          break;
        case series::same:
          of << "#";
          if (t.igat == 0)
            of << "  ";
          else if (t.igat < 10)
            of << t.igat << " ";
          else
            of << t.igat;
          break;
        case series::own:
          of << "%  ";
          break;
      }
      of << setw(wdth-2) << sqrt(t.ga) << ", ";
      if (t.Ps())
      {
        switch(t.fl_gb)
        {
          case series::fixed:
            of << "*  ";
            break;
          case series::same:
            of << "#";
            if (t.igbt == 0)
              of << "  ";
            else if (t.igbt < 10)
              of << t.igbt << " ";
            else
              of << t.igbt;
            break;
          case series::own:
            of << "%  ";
            break;
        }
        of << setw(wdth-2) << sqrt(t.gb);
      }
      else
        of << "       n/a  ";
      of << ';' << endl;
    }
    iser++;
  }
  of << endl;

  of << "     ID         av_dev        sri      err_a        err_b      err1_b" << endl;
  iser = ser.begin();
  while(iser != ser.end())
  {
    series& t = *iser;
    if (t.hide && (print_series == all_series))
      of << '*';
    else if (!t.hide)
      of << ' ';

    if (!t.hide || (print_series == all_series))
    {
      of << setw(wdth) << t.ID;
      of << setw(wdth) << sqrt(t.sumi/t.Ns());
      if (t.Ns() == 1)
        of << setw(wdth) << 0.;
      else if (t.Ps())
        {
          if (t.Ns() == 2)
            of << setw(wdth) << 0.;
          else
          {
          tmp = (t.sumi - t.suma*t.suma/t.Ns() - t.sumb*t.sumb/t.Ps())/t.Ns();
          if (tmp > 0.)
            of << setw(wdth) << sqrt(tmp);
          else
            of << setw(wdth) << 0.;
          }
        }
      else
      {
        tmp = (t.sumi - t.suma*t.suma/t.Ns())/t.Ns();
        if (tmp > 0.)
          of << setw(wdth) << sqrt(tmp);
        else
          of << setw(wdth) << 0.;
      }
      of << setw(wdth) << t.suma/t.Ns();
      if (t.Ps())
        of << setw(wdth) << t.sumb/t.Ps()
           << setw(wdth) << t.sumb/t.Ps()*sqrt(t.Ps()/t.Ns()) << endl;
      else
        of << "       n/a         n/a " << endl;
    }
    iser++;
  }
  of << endl;

  of << "     ID       Ntot     Ns      xav         Ps" << endl;
  iser = ser.begin();
  while(iser != ser.end())
  {
    series& t = *iser;
    if (t.hide && (print_series == all_series))
      of << '*';
    else if (!t.hide)
      of << ' ';

    if (!t.hide || (print_series == all_series))
    {
      of << setw(wdth) << t.ID;
      of << setw(5) << t.Ntot() << setw(5) << t.Ns();
      of << setw(wdth) << t.xs()
         << setw(wdth) << sqrt(t.Ps()) << endl;
    }
    iser++;
  }
  of << endl;
  if (print_series == no_series) return;

  iser = ser.begin();
  while(iser != ser.end())
  {
    int i, j;
    double f;
    series& t = *iser;
    if (!t.hide || (print_series == all_series))
    {
      of << endl;
      for (i = 0; i < t.Nvar()*wdth + 2; i++)
        of << ' ';
      of << " err_full    err_a";
      if (t.Ps())
        of << "   err_ab" << endl;
      else
        of << endl;
    }
    if (t.hide && (print_series == all_series))
      of << '*';
    else if (!t.hide)
      of << ' ';

    if (!t.hide || (print_series == all_series))
    {
      of << setw(wdth) << t.ID << ',' << setw(wdth+2) << t.eq << ',' << endl;
      of << ' ';
      for (j = 0; j < t.Nvar(); j++)
        of << setw(wdth) << t.name(j);
      of << ',' << endl;
      for (i = 0; i < t.Ntot(); i++)
      {
        if (t.atr(i))
          of << ' ';
        else
          of << '*';
        for (j = 0; j < t.Nvar(); j++)
          of << setw(wdth) << t.val(i, j);
        f = (*t.f)(t(i));
        of << setw(wdth) << f;
        if (t.Ns())
        {
          of << setw(wdth) << f - t.suma/t.Ns();
          if (t.Ps())
            of << setw(wdth)
            << f - t.suma/t.Ns() - t.sumb/t.Ps()*(t.val(i, 1) - t.xs());
        }
        if (i < t.Ntot() - 1)
          of << ',' << endl;
        else
          of << ';' << endl;
      }
    }
    iser++;
  }
}
