/*
Copyright (C) 1996-1998 Evgenii Rudnyi, http://Evgenii.Rudnyi.Ru/

Software for paper 
E. B. Rudnyi, V. V. Kuzmenko, G. V. Voronin. Simultaneous assessment of the YBa2Cu6O6+z thermodynamics under the linear error model. J. Phys. Chem. Ref. Data, 1998, v. 27, N 5, p. 855-888.

This software is a copyrighted work licensed under the terms, described
in the file "FREE_LICENSE". 
*/

#ifndef __CUOX_PL_H
#define __CUOX_PL_H

#include "phase.h"

#include "func_tp.h"

class CuOx_plane : public phase
{
  size_t na_;
  size_t nb_;
  vec_func_Tp_ptr fptr_;
  double R_;
  void set_coefs();

public:
  CuOx_plane() : na_(0), nb_(0), R_(8.31441), fptr_(2), phase(2) {}
  CuOx_plane(istream &in) : na_(0), nb_(0), R_(8.31441), fptr_(2), phase(2)
    {read(in);}
  ~CuOx_plane() {}

  virtual void clear()
  {
    na() = 0;
    nb() = 0;
    R_ = 8.31441;
    fptr_.erase(fptr_.begin(), fptr_.end());
    fptr_.push_back(0);
    fptr_.push_back(0);
    phase::clear();
    comps_.push_back(formula());
    comps_.push_back(formula());
  }
  virtual CuOx_plane* create() const {return new CuOx_plane;}
//  virtual CuOx_plane* clone() const {return new CuOx_plane(*this);}
  virtual CuOx_plane* clone() const {return new CuOx_plane;}

  virtual istream& read(istream& in);
  virtual ostream& write(ostream& out) const;

  void reset(size_t i, func_Tp* ptr = 0)
    {fptr_[i].reset(ptr); set_coefs();}
  void reset_ai(size_t i, func_Tp* ptr = 0)
    {fptr_[2 + i].reset(ptr); set_coefs();}
  void reset_bi(size_t i, func_Tp* ptr = 0)
    {fptr_[2 + na() + i].reset(ptr); set_coefs();}
  void push_back_ai(func_Tp* ptr = 0)
    {fptr_.insert(fptr_.begin() + 2 + na()++, ptr); set_coefs();}
  void push_back_bi(func_Tp* ptr = 0)
    {fptr_.insert(fptr_.begin() + 2 + na() + nb()++, ptr); set_coefs();}
  void erase_ai(size_t i)
  {
    if (i < na())
    {
      fptr_.erase(fptr_.begin() + 2 + i);
      --na();
      set_coefs();
    }
  }
  void erase_bi(size_t i)
  {
    if (i < nb())
    {
      fptr_.erase(fptr_.begin() + 2 + na() + i);
      --nb();
      set_coefs();
    }
  }

  size_t& na()
    {return na_;}
  size_t na() const
    {return na_;}
  size_t& nb()
    {return nb_;}
  size_t nb() const
    {return nb_;}
  func_Tp& fi(size_t i)
    {return *fptr_[i];}
  const func_Tp& fi(size_t i) const
    {return *fptr_[i];}
  func_Tp& ai(size_t i)
    {return *fptr_[2 + i];}
  const func_Tp& ai(size_t i) const
    {return *fptr_[2 + i];}
  func_Tp& bi(size_t i)
    {return *fptr_[2 + na() + i];}
  const func_Tp& bi(size_t i) const
    {return *fptr_[2 + na() + i];}

  static void check_z(const double &z, double x = 0.)
  {
    if (z < 0. || z > 1.)
      throw xmsg("z is out_of_range");
//    if (x < -z/2. || x > z/2.)
    if (x && (x < 0. || x > z/2.))
      throw xmsg("x is out_of_range");
  }
  double Z(func_Tp::Z f, const double &T, const double &z, const double &x)
    const;
  double Z(func_Tp::Z f, const double &T, const double &z) const
  {
    double x = x_eq(T, z);
    return Z(f, T, z, x);
  }
  double dZdx(func_Tp::Z f, const double &T, const double &z,
    const double &x) const;
  double dZdz(func_Tp::Z f, const double &T, const double &z,
    const double &x) const;
  double ZO(func_Tp::Z f, const double &T, const double &z,
    const double &x) const
  {
    return dZdz(f, T, z, x) + dZdx(f, T, z, x)*dxdz(T, z, x);
  }
  double ZO(func_Tp::Z f, const double &T, const double &z) const
  {
    double x = x_eq(T, z);
    return dZdz(f, T, z, x) + dZdx(f, T, z, x)*dxdz(T, z, x);
  }
  double x_eq(const double &T, const double &z) const;
  double dxdz(const double &T, const double &z, const double &x) const;
  double dxdT(const double &T, const double &z, const double &x) const;

  double G(const double &T, const double &z, const double &x) const
    {return Z(&func_Tp::G, T, z, x);}
  double G(const double &T, const double &z) const
    {return Z(&func_Tp::G, T, z);}
  double GO(const double &T, const double &z, const double &x) const
    {return ZO(&func_Tp::G, T, z, x);}
  double GO(const double &T, const double &z) const
  {
    double x = x_eq(T, z);
    return dZdz(&func_Tp::G, T, z, x);
  }
  double lnpO2(const double &T, const double &z, const double &x) const
  {
    double g = GO(T, z, x);
    if (g == HUGE_VAL || g == -HUGE_VAL)
      return g;
    return 2.*g/R_/T;
  }
  double lnpO2(const double &T, const double &z) const
  {
    double g = GO(T, z);
    if (g == HUGE_VAL || g == -HUGE_VAL)
      return g;
    return 2.*g/R_/T;
  }
  double H(const double &T, const double &z, const double &x) const
    {return Z(&func_Tp::H, T, z, x);}
  double H(const double &T, const double &z) const
    {return Z(&func_Tp::H, T, z);}
  double HO(const double &T, const double &z, const double &x) const
    {return ZO(&func_Tp::H, T, z, x);}
  double HO(const double &T, const double &z) const
    {return ZO(&func_Tp::H, T, z);}
  double S(const double &T, const double &z, const double &x) const
    {return Z(&func_Tp::S, T, z, x);}
  double S(const double &T, const double &z) const
    {return Z(&func_Tp::S, T, z);}
  double SO(const double &T, const double &z, const double &x) const
    {return ZO(&func_Tp::S, T, z, x);}
  double SO(const double &T, const double &z) const
    {return ZO(&func_Tp::S, T, z);}
  double Cp(const double &T, const double &z, const double &x) const
    {return Z(&func_Tp::Cp, T, z, x);}
  double Cpx(const double &T, const double &z) const
    {return Z(&func_Tp::Cp, T, z);}
  double Cpz(const double &T, const double &z) const
  {
    double x = x_eq(T, z);
    return Cp(T, z, x) + dZdx(&func_Tp::H, T, z, x)*dxdT(T, z, x);
  }

  double z(const double &T, const double &lnpO2, double z_init = 0.5);
  double T_tr(const double &z);
  double T_tr1(const double &lnpO2, double T_tr_init);

  virtual double G(const double& T, const double& p,
          const vec_double &x) const {return G(T, x[1]);}
  virtual double dGdx(const double& T, const double& p,
          const vec_double &x, vec_double &res) const
          {
            double xeq = x_eq(T, x[1]);
            res[0] = 0.;
            res[1] = GO(T, x[1], xeq);
            return G(T, x[1], xeq);
          }
  virtual double mu(const double& T, const double& p,
          const vec_double &x, vec_double &res) const
          {
            double xeq = x_eq(T, x[1]);
            double Zm = G(T, x[1], xeq);
            double dZx = GO(T, x[1], xeq);
            res[0] = Zm - x[1]*dZx;
            res[1] = Zm + (1. - x[1])*dZx;
            return Zm;
          }

  virtual double H(const double& T, const double& p,
          const vec_double &x) const {return H(T, x[1]);}
  virtual double dHdx(const double& T, const double& p,
          const vec_double &x, vec_double &res) const
          {
            double xeq = x_eq(T, x[1]);
            res[0] = 0.;
            res[1] = HO(T, x[1], xeq);
            return H(T, x[1], xeq);
          }
  virtual double h(const double& T, const double& p,
          const vec_double &x, vec_double &res) const
          {
            double xeq = x_eq(T, x[1]);
            double Zm = H(T, x[1], xeq);
            double dZx = HO(T, x[1], xeq);
            res[0] = Zm - x[1]*dZx;
            res[1] = Zm + (1. - x[1])*dZx;
            return Zm;
          }

  virtual double S(const double& T, const double& p,
          const vec_double &x) const {return S(T, x[1]);}
  virtual double dSdx(const double& T, const double& p,
          const vec_double &x, vec_double &res) const
          {
            double xeq = x_eq(T, x[1]);
            res[0] = 0.;
            res[1] = SO(T, x[1], xeq);
            return S(T, x[1], xeq);
          }
  virtual double s(const double& T, const double& p,
          const vec_double &x, vec_double &res) const
          {
            double xeq = x_eq(T, x[1]);
            double Zm = S(T, x[1], xeq);
            double dZx = SO(T, x[1], xeq);
            res[0] = Zm - x[1]*dZx;
            res[1] = Zm + (1. - x[1])*dZx;
            return Zm;
          }

  virtual double Cp(const double& T, const double& p,
          const vec_double &x) const {return Cpz(T, x[1]);}

};

inline istream& operator>> (istream &in, CuOx_plane &to)
{
  in >> ws;
  return to.read(in);
}

inline ostream& operator<< (ostream &out, const CuOx_plane &old)
{
  return old.write(out);
}

#endif
