/*
Copyright (C) 1996-1998 Evgenii Rudnyi, http://Evgenii.Rudnyi.Ru/

Software for paper 
E. B. Rudnyi, V. V. Kuzmenko, G. V. Voronin. Simultaneous assessment of the YBa2Cu6O6+z thermodynamics under the linear error model. J. Phys. Chem. Ref. Data, 1998, v. 27, N 5, p. 855-888.

This software is a copyrighted work licensed under the terms, described
in the file "FREE_LICENSE". 
*/

#ifndef __FUNC_TP_H
#define __FUNC_TP_H

#include <iostream.h>

#include <math.h>

#include <vector.h>

#include "general.h"

#include "auto_ptr1.h"

static const double po = 1.;

class func_Tp : public vec_coef
{
protected:
  vec_string label_;
public:
  typedef double (func_Tp::*Z)(const double &T, const double& p) const;

  func_Tp() {}
  func_Tp(size_t n) : vec_coef(n), label_(n + 1) {}
  virtual ~func_Tp() {}

  virtual void clear()
  {
    erase(begin(), end());
    label_.erase(label_.begin(), label_.end());
  }
  virtual func_Tp* create() const = 0;
  virtual func_Tp* clone() const = 0;

  istream& read(istream& in);
  ostream& write(ostream& out) const;

  vec_string& label() {return label_;}
  const vec_string& label() const {return label_;}
  double& x(size_t i) {return operator[](i).x;}
  const double& x(size_t i) const {return operator[](i).x;}
  bool& unknown(size_t i) {return operator[](i).unknown;}
  const bool& unknown(size_t i) const {return operator[](i).unknown;}
  string& name(size_t i) {return operator[](i).name;}
  const string& name(size_t i) const {return operator[](i).name;}
  string& label(size_t i) {return label_[i];}
  const string& label(size_t i) const {return label_[i];}

  virtual string ClassName() const = 0;
	
  virtual double G(const double &T, const double& p = global::p) const = 0;
  virtual double H(const double &T, const double& p = global::p) const;
  virtual double S(const double &T, const double& p = global::p) const;
  virtual double Cp(const double &T, const double& p = global::p) const;
  virtual double V(const double &T, const double& p = global::p) const;
  virtual double alpha(const double &T, const double& p = global::p) const
    {return 0.;}
  virtual double beta(const double &T, const double& p = global::p) const
    {return 0.;}
};

inline istream& operator>> (istream &in, func_Tp &to)
{
  in >> ws;
  return to.read(in);
}

inline ostream& operator<< (ostream &out, const func_Tp &old)
{
  return old.write(out);
}

extern func_Tp * const NULL_FPTR;

class func_Tp_ptr : public auto_ptr1<func_Tp>
{
  typedef map<string, auto_ptr1<func_Tp>, less<string> > map_types;
  static map_types types;
public:
  func_Tp_ptr(func_Tp *old = 0) : auto_ptr1<func_Tp>(old) {}
  func_Tp_ptr(func_Tp_ptr& old) : auto_ptr1<func_Tp>(old) {}
  func_Tp_ptr(const func_Tp_ptr& old) : auto_ptr1<func_Tp>(*(func_Tp_ptr*)(&old)) {}
  func_Tp_ptr(int, int);
  ~func_Tp_ptr() {}
  static void add(const string& name, func_Tp* ptr)
    {types[name].reset(ptr);}
  func_Tp& operator *() const
  {
    if (get())
      return *get();
    else
      return *NULL_FPTR;
  }
  func_Tp* operator ->() const
  {
    if (get())
      return get();
    else
      return NULL_FPTR;
  }

  istream& read(istream& in);
  ostream& write(ostream& out) const {return (*this)->write(out);}
};

inline istream& operator>> (istream &in, func_Tp_ptr &to)
{
  in >> ws;
  return to.read(in);
}

inline ostream& operator<< (ostream &out, const func_Tp_ptr &old)
{
  return old.write(out);
}

typedef vector<func_Tp_ptr> vec_func_Tp_ptr;

class null_func_Tp : public func_Tp
{
public:
  null_func_Tp() : func_Tp(0) {label_[0] = "(0)";}
  virtual null_func_Tp* create() const {return new null_func_Tp;}
  virtual null_func_Tp* clone() const {return new null_func_Tp(*this);}
  virtual string ClassName() const {return "null_func_Tp";};
  virtual double G(const double &T, const double &p = global::p) const {return 0.;}
  virtual double H(const double &T, const double &p = global::p) const {return 0.;}
  virtual double S(const double &T, const double &p = global::p) const {return 0.;}
  virtual double Cp(const double &T, const double &p = global::p) const {return 0.;}
  virtual double V(const double &T, const double &p = global::p) const {return 0.;}
};

class Cp_zero : public func_Tp
{
public:
  Cp_zero() : func_Tp(2)
  {
    label(0) = "(";
    name(0) = "a";
    label(1) = "+";
    name(1) = "b";
    label(2) = "*T)";
  }
  virtual Cp_zero* create() const {return new Cp_zero;}
  virtual Cp_zero* clone() const {return new Cp_zero(*this);}
  virtual string ClassName() const {return "Cp_zero";};
  virtual double G(const double &T, const double &p = global::p) const
    {return x(0) + T*x(1);}
  virtual double H(const double &T, const double &p = global::p) const
    {return x(0);}
  virtual double S(const double &T, const double &p = global::p) const
    {return -x(1);}
  virtual double Cp(const double &T, const double &p = global::p) const
    {return 0.;}
  virtual double V(const double &T, const double &p = global::p) const
    {return 0.;}
};

class Cp_zero_V : public func_Tp
{
public:
  Cp_zero_V() : func_Tp(3)
  {
    label(0) = "(";
    name(0) = "a";
    label(1) = "+";
    name(1) = "b";
    label(2) = "*T+";
    name(2) = "V";
    label(3) = "*p)";
  }
  virtual Cp_zero_V* create() const {return new Cp_zero_V;}
  virtual Cp_zero_V* clone() const {return new Cp_zero_V(*this);}
  virtual string ClassName() const {return "Cp_zero_V";};
  virtual double G(const double &T, const double &p = global::p) const
    {return x(0) + T*x(1) + p*x(2);}
  virtual double H(const double &T, const double &p = global::p) const
    {return x(0) + p*x(2);}
  virtual double S(const double &T, const double &p = global::p) const
    {return -x(1);}
  virtual double Cp(const double &T, const double &p = global::p) const
    {return 0.;}
  virtual double V(const double &T, const double &p = global::p) const
    {return x(2);}
};

class Cp_const : public func_Tp
{
public:
  Cp_const() : func_Tp(3)
  {
    label(0) = "(";
    name(0) = "a";
    label(1) = "+";
    name(1) = "b";
    label(2) = "*T+";
    name(2) = "c";
    label(3) = "*T*log(T))";
  }
  virtual Cp_const* create() const {return new Cp_const;}
  virtual Cp_const* clone() const {return new Cp_const(*this);}
  virtual string ClassName() const {return "Cp_const";};
  virtual double G(const double &T, const double &p = global::p) const
    {return x(0) + T*x(1) + T*log(T)*x(2);}
  virtual double H(const double &T, const double &p = global::p) const
    {return x(0) - T*x(2);}
  virtual double S(const double &T, const double &p = global::p) const
    {return -(x(1) + (log(T)+1.)*x(2));}
  virtual double Cp(const double &T, const double &p = global::p) const
    {return -x(2);}
  virtual double V(const double &T, const double &p = global::p) const
    {return 0.;}
};

class Cp_BB2 : public func_Tp
{
public:
  Cp_BB2() : func_Tp(4)
  {
    label(0) = "(";
    name(0) = "a";
    label(1) = "+";
    name(1) = "b";
    label(2) = "*T+";
    name(2) = "c";
    label(3) = "*T*log(T)+";
    name(3) = "d";
    label(4) = "*sqrt(T))";
  }
  virtual Cp_BB2* create() const {return new Cp_BB2;}
  virtual Cp_BB2* clone() const {return new Cp_BB2(*this);}
  virtual string ClassName() const {return "Cp_BB2";};
  virtual double G(const double &T, const double &p = global::p) const
    {return x(0) + T*x(1) + T*log(T)*x(2) + sqrt(T)*x(3);}
  virtual double H(const double &T, const double &p = global::p) const
    {return x(0) - T*x(2) + 0.5*sqrt(T)*x(3);}
  virtual double S(const double &T, const double &p = global::p) const
    {return -(x(1) + (log(T)+1.)*x(2) + 0.5/sqrt(T)*x(3));}
  virtual double Cp(const double &T, const double &p = global::p) const
    {return -x(2) + 0.25/sqrt(T)*x(3);}
  virtual double V(const double &T, const double &p = global::p) const
    {return 0.;}
};

class Cp_BB2_V : public func_Tp
{
public:
  Cp_BB2_V() : func_Tp(5)
  {
    label(0) = "(";
    name(0) = "a";
    label(1) = "+";
    name(1) = "b";
    label(2) = "*T+";
    name(2) = "c";
    label(3) = "*T*log(T)+";
    name(3) = "d";
    label(4) = "*sqrt(T)+";
    name(4) = "V";
    label(5) = "*p)";
  }
  virtual Cp_BB2_V* create() const {return new Cp_BB2_V;}
  virtual Cp_BB2_V* clone() const {return new Cp_BB2_V(*this);}
  virtual string ClassName() const {return "Cp_BB2_V";};
  virtual double G(const double &T, const double &p = global::p) const
    {return x(0) + T*x(1) + T*log(T)*x(2) + sqrt(T)*x(3) + p*x(4);}
  virtual double H(const double &T, const double &p = global::p) const
    {return x(0) - T*x(2) + 0.5*sqrt(T)*x(3) + p*x(4);}
  virtual double S(const double &T, const double &p = global::p) const
    {return -(x(1) + (log(T)+1.)*x(2) + 0.5/sqrt(T)*x(3));}
  virtual double Cp(const double &T, const double &p = global::p) const
    {return -x(2) + 0.25/sqrt(T)*x(3);}
  virtual double V(const double &T, const double &p = global::p) const
    {return x(4);}
};

class Cp_BB2_V2 : public func_Tp
{
public:
  Cp_BB2_V2() : func_Tp(6)
  {
    label(0) = "(";
    name(0) = "a";
    label(1) = "+";
    name(1) = "b";
    label(2) = "*T+";
    name(2) = "c";
    label(3) = "*T*log(T)+";
    name(3) = "d";
    label(4) = "*sqrt(T)+";
    name(4) = "Vo";
    label(5) = "*exp(";
    name(5) = "alpha";
    label(6) = "*T)*(p-po)";
  }
  virtual Cp_BB2_V2* create() const {return new Cp_BB2_V2;}
  virtual Cp_BB2_V2* clone() const {return new Cp_BB2_V2(*this);}
  virtual string ClassName() const {return "Cp_BB2_V2";};
  virtual double G(const double &T, const double &p = global::p) const
    {return x(0) + T*x(1) + T*log(T)*x(2) + sqrt(T)*x(3) +
            x(4)*exp(x(5)*T)*(p - po);}
  virtual double H(const double &T, const double &p = global::p) const
    {return x(0) - T*x(2) + 0.5*sqrt(T)*x(3) +
            x(4)*exp(x(5)*T)*(p - po)*(1. - x(5)*T);}
  virtual double S(const double &T, const double &p = global::p) const
    {return -(x(1) + (log(T)+1.)*x(2) + 0.5/sqrt(T)*x(3) +
            x(4)*x(5)*exp(x(5)*T)*(p - po));}
  virtual double Cp(const double &T, const double &p = global::p) const
    {return -x(2) + 0.25/sqrt(T)*x(3) -
            x(4)*exp(x(5)*T)*(p - po)*x(5)*x(5)*T;}
  virtual double V(const double &T, const double &p = global::p) const
    {return x(4)*exp(x(5)*T);}
  virtual double alpha(const double &T, const double &p = global::p) const
    {return x(5);}
};

class Cp_BB2_V3 : public func_Tp
{
public:
  Cp_BB2_V3() : func_Tp(7)
  {
    label(0) = "(";
    name(0) = "a";
    label(1) = "+";
    name(1) = "b";
    label(2) = "*T+";
    name(2) = "c";
    label(3) = "*T*log(T)+";
    name(3) = "d";
    label(4) = "*sqrt(T)+";
    name(4) = "Vb";
    label(5) = "*exp(";
    name(5) = "alpha";
    label(6) = "*T)*(1.-exp(";
    name(6) = "beta";
    label(7) = "*(po-p)))";
  }
  virtual Cp_BB2_V3* create() const {return new Cp_BB2_V3;}
  virtual Cp_BB2_V3* clone() const {return new Cp_BB2_V3(*this);}
  virtual string ClassName() const {return "Cp_BB2_V3";};
  virtual double G(const double &T, const double &p = global::p) const
    {return x(0) + T*x(1) + T*log(T)*x(2) + sqrt(T)*x(3) +
            x(4)*exp(x(5)*T)*(1. - exp(x(6)*(po - p)));}
  virtual double H(const double &T, const double &p = global::p) const
    {return x(0) - T*x(2) + 0.5*sqrt(T)*x(3) +
            x(4)*exp(x(5)*T)*(1. - exp(x(6)*(po - p)))*(1. - x(5)*T);}
  virtual double S(const double &T, const double &p = global::p) const
    {return -(x(1) + (log(T)+1.)*x(2) + 0.5/sqrt(T)*x(3) +
            x(4)*x(5)*exp(x(5)*T)*(1. - exp(x(6)*(po - p))));}
  virtual double Cp(const double &T, const double &p = global::p) const
    {return -x(2) + 0.25/sqrt(T)*x(3) -
            x(4)*exp(x(5)*T)*(1. - exp(x(6)*(po - p)))*x(5)*x(5)*T;}
  virtual double V(const double &T, const double &p = global::p) const
    {return x(4)*x(6)*exp(x(5)*T)*exp(x(6)*(po - p));}
  virtual double alpha(const double &T, const double &p = global::p) const
    {return x(5);}
  virtual double beta(const double &T, const double &p = global::p) const
    {return x(6);}
};

class Cp_BB2_V3_ : public func_Tp
{
public:
  Cp_BB2_V3_() : func_Tp(7)
  {
    label(0) = "(";
    name(0) = "a";
    label(1) = "+";
    name(1) = "b";
    label(2) = "*T+";
    name(2) = "c";
    label(3) = "*T*log(T)+";
    name(3) = "d";
    label(4) = "*sqrt(T)+";
    name(4) = "V";
    label(5) = "/beta*exp(";
    name(5) = "alpha";
    label(6) = "*T)*(1.-exp(";
    name(6) = "beta";
    label(7) = "*(po-p)))";
  }
  virtual Cp_BB2_V3_* create() const {return new Cp_BB2_V3_;}
  virtual Cp_BB2_V3_* clone() const {return new Cp_BB2_V3_(*this);}
  virtual string ClassName() const {return "Cp_BB2_V3_";};
  virtual double G(const double &T, const double &p = global::p) const
  {
    if (x(6))
      return x(0) + T*x(1) + T*log(T)*x(2) + sqrt(T)*x(3) +
             x(4)/x(6)*exp(x(5)*T)*(1. - exp(x(6)*(po - p)));
    else
      return x(0) + T*x(1) + T*log(T)*x(2) + sqrt(T)*x(3) +
             x(4)*exp(x(5)*T)*(p - po);
  }
  virtual double H(const double &T, const double &p = global::p) const
  {
    if (x(6))
      return x(0) - T*x(2) + 0.5*sqrt(T)*x(3) +
             x(4)/x(6)*exp(x(5)*T)*(1. - exp(x(6)*(po - p)))*(1. - x(5)*T);
    else
      return x(0) - T*x(2) + 0.5*sqrt(T)*x(3) +
             x(4)*exp(x(5)*T)*(p - po)*(1. - x(5)*T);
  }
  virtual double S(const double &T, const double &p = global::p) const
  {
    if (x(6))
      return -(x(1) + (log(T)+1.)*x(2) + 0.5/sqrt(T)*x(3) +
             x(4)/x(6)*x(5)*exp(x(5)*T)*(1. - exp(x(6)*(po - p))));
    else
      return -(x(1) + (log(T)+1.)*x(2) + 0.5/sqrt(T)*x(3) +
             x(4)*x(5)*exp(x(5)*T)*(p - po));
  }
  virtual double Cp(const double &T, const double &p = global::p) const
  {
    if (x(6))
      return -x(2) + 0.25/sqrt(T)*x(3) -
             x(4)/x(6)*exp(x(5)*T)*(1. - exp(x(6)*(po - p)))*x(5)*x(5)*T;
    else
      return -x(2) + 0.25/sqrt(T)*x(3) -
             x(4)*exp(x(5)*T)*(p - po)*x(5)*x(5)*T;
  }
  virtual double V(const double &T, const double &p = global::p) const
    {return x(4)*exp(x(5)*T)*exp(x(6)*(po - p));}
  virtual double alpha(const double &T, const double &p = global::p) const
    {return x(5);}
  virtual double beta(const double &T, const double &p = global::p) const
    {return x(6);}
};

class IVT_Tp : public func_Tp
{
public:
  IVT_Tp() : func_Tp(7)
  {
    label(0) = "(";
    name(0) = "a";
    label(1) = "+";
    name(1) = "b";
    label(2) = "*T+";
    name(2) = "c";
    label(3) = "*T*log(T)+";
    name(3) = "d";
    label(4) = "*T^2+";
    name(4) = "e";
    label(5) = "*T^3+";
    name(5) = "f";
    label(6) = "*T^4+";
    name(6) = "g";
    label(7) = "/T)";
  }
  virtual IVT_Tp* create() const {return new IVT_Tp;}
  virtual IVT_Tp* clone() const {return new IVT_Tp(*this);}

  virtual string ClassName() const {return "IVT_Tp";};
// g1 := x0 + x1*Te + x2*Te*log(Te) + x3*Te^2 + x4*Te^3 + x5*Te^4 + x6/Te;
  virtual double G(const double &T, const double &p = global::p) const
    {return x(0) + x(1)*T + x(2)*T*log(T)
                 + x(3)*pow(T, 2) + x(4)*pow(T, 3)
                 + x(5)*pow(T, 4) + x(6)/T;}

//h1 := g1 + Te*s1;
//              4          3        2                  -1
//H1 :=  - (3*TE *X5 + 2*TE *X4 + TE *X3 + TE*X2 - 2*TE  *X6 - X0)
//
//-Te^2*df(g1/Te,Te);
//        4          3        2                  -1
// - (3*TE *X5 + 2*TE *X4 + TE *X3 + TE*X2 - 2*TE  *X6 - X0)
  virtual double H(const double &T, const double &p = global::p) const
    {return x(0) - x(2)*T - x(3)*pow(T, 2)
                 - x(4)*2.*pow(T, 3)
                 - x(5)*3.*pow(T, 4)
                 + x(6)*2./T;}

//s1 := -df(g1, te);
//                           3          2                  -2
//S1 :=  - (LOG(TE)*X2 + 4*TE *X5 + 3*TE *X4 + 2*TE*X3 - TE  *X6 + X1 +
//
//           X2)
  virtual double S(const double &T, const double &p = global::p) const
    {return -(x(1) + x(2)*(log(T)+1.)
                   + x(3)*2.*T
                   + x(4)*3.*pow(T, 2)
                   + x(5)*4.*pow(T, 3)
                   - x(6)/pow(T, 2));}

//cp1 := df(h1, Te);
//                3          2                    -2
//CP1 :=  - (12*TE *X5 + 6*TE *X4 + 2*TE*X3 + 2*TE  *X6 + X2)
//
//Te*df(s1, Te);
//         3          2                    -2
// - (12*TE *X5 + 6*TE *X4 + 2*TE*X3 + 2*TE  *X6 + X2)
  virtual double Cp(const double &T, const double &p = global::p) const
    {return -(x(2) + x(3)*2.*T
                   + x(4)*6.*pow(T, 2)
                   + x(5)*12.*pow(T, 3)
                   + x(6)*2./pow(T, 2));}
  virtual double V(const double &T, const double &p = global::p) const
    {return 0.;}
};

class Cp_BB4 : public func_Tp
{
public:
  Cp_BB4() : func_Tp(6)
  {
    label(0) = "(";
    name(0) = "a";
    label(1) = "+";
    name(1) = "b";
    label(2) = "*T+";
    name(2) = "c";
    label(3) = "*T*log(T)+";
    name(3) = "d";
    label(4) = "*sqrt(T)+";
    name(4) = "e";
    label(5) = "/T+";
    name(5) = "f";
    label(6) = "/T^2)";
  }
  virtual Cp_BB4* create() const {return new Cp_BB4;}
  virtual Cp_BB4* clone() const {return new Cp_BB4(*this);}

  virtual string ClassName() const {return "Cp_BB4";};
// g := x0 + x1*Te + x2*Te*log(Te) + x3*sqrt(Te) + x4/Te + x5/Te^2;
  virtual double G(const double &T, const double &p = global::p) const
    {return x(0) + T*x(1) + T*log(T)*x(2) + sqrt(T)*x(3) +
            x(4)/T + x(5)*pow(T, -2);}

//h := g + Te*s;
//             1          -1                     -1          -2
//H :=  - ( - ---*SQRT(TE)  *TE*X3 + TE*X2 - 2*TE  *X4 - 3*TE  *X5 - X0)
//             2
// -Te^2*df(g/Te,Te);
//        1          -1                     -1          -2
// - ( - ---*SQRT(TE)  *TE*X3 + TE*X2 - 2*TE  *X4 - 3*TE  *X5 - X0)
//        2
  virtual double H(const double &T, const double &p = global::p) const
    {return x(0) - T*x(2) + 0.5*sqrt(T)*x(3) +
            2.*x(4)/T + 3.*x(5)*pow(T, -2);}

//s := -df(g, te);
//          1          -1                     -2          -3
//S :=  - (---*SQRT(TE)  *X3 + LOG(TE)*X2 - TE  *X4 - 2*TE  *X5 + X1 +  X2)
//          2
  virtual double S(const double &T, const double &p = global::p) const
    {return -(x(1) + (log(T)+1.)*x(2) + 0.5/sqrt(T)*x(3)
            -x(4)*pow(T, -2) - 2.*x(5)*pow(T, -3));}

//cp := df(h, Te);
//              1          -1          -2          -3
//CP :=  - ( - ---*SQRT(TE)  *X3 + 2*TE  *X4 + 6*TE  *X5 + X2)
//              4
//Te*df(s, Te);
//        1          -1          -2          -3
// - ( - ---*SQRT(TE)  *X3 + 2*TE  *X4 + 6*TE  *X5 + X2)
//        4
  virtual double Cp(const double &T, const double &p = global::p) const
    {return -x(2) + 0.25/sqrt(T)*x(3)
            - 2.0*x(4)*pow(T, -2) - 6.*x(5)*pow(T, -3);}
  virtual double V(const double &T, const double &p = global::p) const
    {return 0.;}
};

#endif
