/*
Copyright (C) 1996-1998 Evgenii Rudnyi, http://Evgenii.Rudnyi.Ru/

Software for paper 
E. B. Rudnyi, V. V. Kuzmenko, G. V. Voronin. Simultaneous assessment of the YBa2Cu6O6+z thermodynamics under the linear error model. J. Phys. Chem. Ref. Data, 1998, v. 27, N 5, p. 855-888.

This software is a copyrighted work licensed under the terms, described
in the file "FREE_LICENSE". 
*/

#include "phase.h"
#include <math.h>
#include <float.h>

void phase::check_x(const vec_double &x) const
{
  if (set_check_x_)
  {
    if (x.size() != comps_.size())
      throw xmsg("mole fraction vector is uncompatible with the phase");
    double s = 0.;
    for (int i = 0; i != x.size(); ++i)
    {
      if (x[i] < 0.)
        throw xmsg("negative mole fraction");
      s += x[i];
    }
    if (fabs(s - 1.) > DBL_EPSILON*10.)
      throw xmsg("sum of mole fractions is not unit");
  }
}

double phase::dGdx(const double& T, const double& p,
                   const vec_double &x, vec_double &res) const
{
  double Gc = G(T, p, x);
  vec_double x1 = x;
  double stepx;
  for (int i = 0; i != comps_.size(); ++i)
  {
    stepx = global::step1*(x1[i] + global::step1);
    if (x1[i] + stepx > 1.) stepx = -stepx;
    x1[i] += stepx;
    stepx = x1[i] - x[i];
    res[i] = (G(T, p, x1) - Gc)/stepx;
    x1[i] = x[i];
  }
  return Gc;
}

double phase::mu(const double& T, const double& p,
                 const vec_double &x, vec_double &res) const
{
  double Z = dGdx(T, p, x, res);
  double Sum = Z;
  int i;
  for (i = 0; i != comps_.size(); ++i)
    Sum -= res[i]*x[i];
  for (i = 0; i != comps_.size(); ++i)
    res[i] += Sum;
  return Z;
}

double phase::H(const double& T, const double& p, const vec_double &x) const
{
  double Gc = G(T, p, x);
  double stepT = global::step1*(T + global::step1);
  double T1 = T + stepT;
  stepT = T1 - T;
  double Sn = -(G(T1, p, x) - Gc)/stepT;
  return Gc + T*Sn;
}

double phase::dHdx(const double& T, const double& p,
                   const vec_double &x, vec_double &res) const
{
  double Gc = G(T, p, x);
  double stepT = global::step2*(T + global::step2);
  double T1 = T + stepT;
  stepT = T1 - T;
  double GT = G(T1, p, x);
  vec_double x1 = x;
  double stepx, dSx, dGx, Gx;
  for (int i = 0; i != comps_.size(); ++i)
  {
    stepx = global::step2*(x1[i] + global::step2);
    if (x1[i] + stepx > 1.) stepx = -stepx;
    x1[i] += stepx;
    stepx = x1[i] - x[i];
    Gx = G(T, p, x1);
    dGx = (Gx - Gc)/stepx;
    dSx = -(G(T1, p, x1) - GT - Gx + Gc)/stepT/stepx;
    res[i] = dGx + T*dSx;
    x1[i] = x[i];
  }
  double Sn = -(GT - Gc)/stepT;
  return Gc + T*Sn;
}

double phase::h(const double& T, const double& p,
                const vec_double &x, vec_double &res) const
{
  double Z = dHdx(T, p, x, res);
  double Sum = Z;
  int i;
  for (i = 0; i != comps_.size(); ++i)
    Sum -= res[i]*x[i];
  for (i = 0; i != comps_.size(); ++i)
    res[i] += Sum;
  return Z;
}

double phase::S(const double& T, const double& p, const vec_double &x) const
{
  double Gc = G(T, p, x);
  double stepT = global::step1*(T + global::step1);
  double T1 = T + stepT;
  stepT = T1 - T;
  return -(G(T1, p, x) - Gc)/stepT;
}

double phase::dSdx(const double& T, const double& p,
                const vec_double &x, vec_double &res) const
{
  double Gc = G(T, p, x);
  double stepT = global::step2*(T + global::step2);
  double T1 = T + stepT;
  stepT = T1 - T;
  double GT = G(T1, p, x);
  vec_double x1 = x;
  double stepx;
  for (int i = 0; i != comps_.size(); ++i)
  {
    stepx = global::step2*(x1[i] + global::step2);
    if (x1[i] + stepx > 1.) stepx = -stepx;
    x1[i] += stepx;
    stepx = x1[i] - x[i];
    res[i] = -(G(T1, p, x1) - GT - G(T, p , x1) + Gc)/stepT/stepx;
    x1[i] = x[i];
  }
  return -(GT - Gc)/stepT;
}

double phase::s(const double& T, const double& p,
                const vec_double &x, vec_double &res) const
{
  double Z = dHdx(T, p, x, res);
  double Sum = Z;
  int i;
  for (i = 0; i != comps_.size(); ++i)
    Sum -= res[i]*x[i];
  for (i = 0; i != comps_.size(); ++i)
    res[i] += Sum;
  return Z;
}

double phase::Cp(const double& T, const double& p, const vec_double &x) const
{
  double stepT = global::step2*(T + global::step2);
  double T1 = T - stepT;
  double T2 = T + stepT;
  double df = (G(T1, p, x) - 2.*G(T, p, x) + G(T2, p, x))/stepT/stepT;
  return -T*df;
}

