/*
Copyright (C) 1996-1998 Evgenii Rudnyi, http://Evgenii.Rudnyi.Ru/

Software for paper 
E. B. Rudnyi, V. V. Kuzmenko, G. V. Voronin. Simultaneous assessment of the YBa2Cu6O6+z thermodynamics under the linear error model. J. Phys. Chem. Ref. Data, 1998, v. 27, N 5, p. 855-888.

This software is a copyrighted work licensed under the terms, described
in the file "FREE_LICENSE". 
*/

#ifndef __PHASE_H
#define __PHASE_H

#include "general.h"

class phase
{
public:
  class const_iterator;
  class iterator : public forward_iterator<coefficient, ptrdiff_t>
  {
    vec_coef_ptr::iterator ptr;
    friend class const_iterator;
  public:
    iterator(vec_coef_ptr::iterator ptr_) : ptr(ptr_) {}
    iterator() {}
    bool operator==(const iterator& y) const { return ptr == y.ptr;}
    coefficient& operator*() {return **ptr;}
    iterator& operator++() {++ptr; return *this;}
    iterator operator++(int)
    {
      iterator tmp = *this;
      ++*this;
      return tmp;
    }
  };

  class const_iterator : public forward_iterator<coefficient, ptrdiff_t>
  {
    vec_coef_ptr::const_iterator ptr;
  public:
    const_iterator(vec_coef_ptr::const_iterator ptr_) : ptr(ptr_) {}
    const_iterator(const iterator y) : ptr(y.ptr) {}
    const_iterator() {}
    bool operator==(const const_iterator& y) const { return ptr == y.ptr;}
    const coefficient& operator*() {return **ptr;}
    const_iterator& operator++() {++ptr; return *this;}
    const_iterator operator++(int)
    {
      const_iterator tmp = *this;
      ++*this;
      return tmp;
    }
  };

protected:
  string name_;
  vec_formula comps_;
  bool set_check_x_;
  vec_coef_ptr coefs_;

public:
  phase() {set_check_x_ = false;}
  phase(size_t n) : comps_(n) {set_check_x_ = false;}
  virtual ~phase() {}

  virtual void clear()
  {
    name_.clear();
    comps_.erase(comps_.begin(), comps_.end());
    set_check_x_ = false;
    coefs_.erase(coefs_.begin(), coefs_.end());
  }
  virtual phase* create() const = 0;
  virtual phase* clone() const = 0;

  virtual istream& read(istream& in) = 0;
  virtual ostream& write(ostream& out) const = 0;

  string& name() {return name_;}
  const string& name() const {return name_;}
  vec_formula& comps() {return comps_;}
  const vec_formula& comps() const {return comps_;}

  bool& set_check_x() {return set_check_x_;}
  bool set_check_x() const {return set_check_x_;}
  void check_x(const vec_double &x) const;

  iterator begin() {return coefs_.begin();}
  iterator end() {return coefs_.end();}
  const_iterator begin() const {return coefs_.begin();}
  const_iterator end() const {return coefs_.end();}
  size_t size() const {return coefs_.size();}
  bool empty() const {return coefs_.empty();}

  virtual double G(const double& T, const double& p,
                   const vec_double &x) const = 0;
  virtual double dGdx(const double& T, const double& p,
                    const vec_double &x, vec_double &res) const;
  virtual double mu(const double& T, const double& p,
                    const vec_double &x, vec_double &res) const;

  virtual double H(const double& T, const double& p,
                   const vec_double &x) const;
  virtual double dHdx(const double& T, const double& p,
                    const vec_double &x, vec_double &res) const;
  virtual double h(const double& T, const double& p,
                 const vec_double &x, vec_double &res) const;

  virtual double S(const double& T, const double& p,
                   const vec_double &x) const;
  virtual double dSdx(const double& T, const double& p,
                    const vec_double &x, vec_double &res) const;
  virtual double s(const double& T, const double& p,
                 const vec_double &x, vec_double &res) const;

  virtual double Cp(const double& T, const double& p,
                 const vec_double &x) const;

  double G(const double& T, const vec_double &x) const
    {return G(T, global::p, x);}
  double dGdx(const double& T, const vec_double &x, vec_double &res) const
    {return dGdx(T, global::p, x, res);}
  double mu(const double& T, const vec_double &x, vec_double &res) const
    {return mu(T, global::p, x, res);}

  double H(const double& T, const vec_double &x) const
    {return H(T, global::p, x);}
  double dHdx(const double& T, const vec_double &x, vec_double &res) const
    {return dHdx(T, global::p, x, res);}
  double h(const double& T, const vec_double &x, vec_double &res) const
    {return h(T, global::p, x, res);}

  double S(const double& T, const vec_double &x) const
    {return S(T, global::p, x);}
  double dSdx(const double& T, const vec_double &x, vec_double &res) const
    {return dSdx(T, global::p, x, res);}
  double s(const double& T, const vec_double &x, vec_double &res) const
    {return s(T, global::p, x, res);}

  double Cp(const double& T, const vec_double &x) const
    {return Cp(T, global::p, x);}
};

inline istream& operator>>(istream &in, phase &to)
{
  in >> ws;
  return to.read(in);
};

inline ostream& operator<<(ostream &out, const phase &old)
{
  return old.write(out);
};

#endif

