/*
Copyright (C) 1996-1998 Evgenii Rudnyi, http://Evgenii.Rudnyi.Ru/

Software for paper 
E. B. Rudnyi, V. V. Kuzmenko, G. V. Voronin. Simultaneous assessment of the YBa2Cu6O6+z thermodynamics under the linear error model. J. Phys. Chem. Ref. Data, 1998, v. 27, N 5, p. 855-888.

This software is a copyrighted work licensed under the terms, described
in the file "FREE_LICENSE". 
*/

#include <fstream.h>
#include <iomanip.h>
#include <string.h>
#include "cuox_pl.h"

const char *HELP =
"\nEvgenii Rudnyi (C) All rights reserved, 1996 \n\n\
usage:\n\n\
\t\tY123 -T z Tmin Tmax Tstep\n\
or\n\
\t\tY123 -z T zmin zmax zstep\n\n";

double R = 8.31441;
func_Tp_ptr Y2O3, BaO, CuO, O2;
CuOx_plane Y123;
double p = 1.;

void print_all(const CuOx_plane &Y123, const double &T, const double &z)
{
  double x = Y123.x_eq(T, z);
  double Tr = 298.15;
  double H298 = Y123.H(Tr, z)*R +
  (0.5*Y2O3->H(Tr) + 2*BaO->H(Tr) + 3*CuO->H(Tr) + (z-0.5)/2.*O2->H(Tr));
  cout << setw(8) << T << setw(8) << z << setw(8) << x
//       << setw(10) << Y123.Cp(T, z, x)*R +
//  (0.5*Y2O3->Cp(T) + 2*BaO->Cp(T) + 3*CuO->Cp(T) + (z-0.5)/2.*O2->Cp(T))
       << setw(10) << Y123.Cpz(T, z)*R +
  (0.5*Y2O3->Cp(T) + 2*BaO->Cp(T) + 3*CuO->Cp(T) + (z-0.5)/2.*O2->Cp(T))
       << setw(10) << Y123.S(T, z, x)*R +
  (0.5*Y2O3->S(T) + 2*BaO->S(T) + 3*CuO->S(T) + (z-0.5)/2.*O2->S(T))
       << setw(10) << (Y123.H(T, z, x)*R +
  (0.5*Y2O3->H(T) + 2*BaO->H(T) + 3*CuO->H(T) + (z-0.5)/2.*O2->H(T))
  - H298)/1000.
       << setw(10) << Y123.H(T, z, x)*R/1000.;
     double tmp = Y123.lnpO2(T, z, x);
     if (tmp > 1e6)
       cout << setw(10)<< "inf";
     else if (tmp < -1e6)
       cout << setw(10) << "-inf";
     else
       cout << setw(10) << tmp;

//       << setw(12) << Y123.G(T, z, x)*R
//       << setw(12) << Y123.H(T, z, x)*R
//       << setw(12) << Y123.S(T, z, x)*R
//       << setw(12) << Y123.Cpx(T, z)*R
//       << setw(12) << Y123.Cpz(T, z)*R
//       << setw(12) << Y123.lnpO2(T, z, x)
//       << setw(12) << Y123.HO(T, z, x)*R
//       << setw(12) << Y123.SO(T, z, x)
       cout << endl;

}

int main(int argc, char *argv[])
{
	try {
  cout.precision(2);
  cout.setf(ios::showpoint);
  cout.setf(ios::fixed, ios::floatfield);
  ifstream in("y123.mod");
  in >> Y123 >> Y2O3 >> BaO >> CuO >> O2;
  if (argc < 3 || argc > 6)
  {
    cout << HELP;
    return 0;
  }
//  in >> Y123;
//  cout << Y123;
//  cout <<
//"      T        z         x           G           H           S          Cpx        Cpz        lnpO2          HO         SO"
//    << endl;
//  p = atof(argv[1]);
//  cout << "p= " << p << endl;
  cout <<
"    T        z       x        Cpz        S      H - H298  Del_ox_H ln{pO2/po}"
    << endl;

  if (strcmp(argv[1], "-T") == 0)
  {
    double z = 0., Tmin = 0., Tmax = 0., Tstep = 0.;
    z = atof(argv[2]);
    CuOx_plane::check_z(z);
    if (argc > 3)
      Tmin = atof(argv[3]);
    if (argc > 4)
      Tmax = atof(argv[4]);
    if (argc > 5)
      Tstep = atof(argv[5]);
    if (Tmin < 100.) Tmin = 100.;
    if (Tmax == 0.)
      Tmax = Tmin;
    else if (Tmax < 100.)
      Tmax = 100.;
    if (Tstep == 0.) Tstep = (Tmax - Tmin)/21.;
    if (Tstep == 0.) Tstep = 1.;
    if ((Tmax - Tmin)*Tstep < 0.) Tstep = -Tstep;
    for (double T = Tmin;
         Tstep >=0 ? T <= Tmax + Tstep/2. : T >= Tmax - Tstep/2.;
         T += Tstep)
      print_all(Y123, T, z);
    return 0;
  }
  else if (strcmp(argv[1], "-z") == 0)
  {
    double T = 0., zmin = 0., zmax = -1., zstep = 0.;
    T = atof(argv[2]);
    if (T < 100)
    {
      cout << HELP;
      return 1;
    }
    if (argc > 3)
      zmin = atof(argv[3]);
    if (argc > 4)
      zmax = atof(argv[4]);
    if (argc > 5)
      zstep = atof(argv[5]);
    if (zmin < 0.) zmin = 0.;
    if (zmax == -1.)
      zmax = 1.;
    else if (zmax < 0.)
      zmax = 0.;
    if (zmax > 1.) zmax = 1.;
    if (zstep == 0.) zstep = (zmax - zmin)/21.;
    if (zstep == 0.) zstep = 1.;
    if ((zmax - zmin)*zstep < 0.) zstep = -zstep;
    CuOx_plane::check_z(zmin);
    CuOx_plane::check_z(zmax);
    for (double z = zmin;
         zstep >=0 ? z <= zmax + zstep/2. : z >= zmax - zstep/2.;
         z += zstep)
    {
      if (z > 1.) z = 1.;
      print_all(Y123, T, z);
    }
    return 0;
  }
  else
  {
    cout << HELP;
    return 0;
  }
	}
	catch (xmsg t)
	{
		cout << "There was an exception xmsg: " << t.print() << endl;
	}
	catch(...)
	{
		cout << "There was anknown exception" << endl;
	}
}
